using System;

namespace Printy.Library
{
	/// <summary>
	/// Knoten des ATG-Formattierungs-Syntaxbaums.
	/// </summary>
  public class Node
  {
    /// <summary>
    /// Codes fr den Typ eines Knotens
    /// </summary>
    public const int
      Grammar    = 0,   // Grammatik (Menge von Produktionen + Prolog und Epilog)
      Log        = 1,   // Prolog oder Epilog
      Production = 2,   // Eine Produktion
      Symbol     = 3,   // Ein Symbol (TS oder NTS).
      Literal    = 4,   // Literal (string | char | "WEAK" | "SYNC" | "ANY")
      Group      = 5,   // Klammerung () oder [] oder {}
      Resolver   = 6,   // LL(1) conflict resolver.
      Action     = 7,   // Semantische Aktion
      Attribute  = 8,   // Attribut(e)
      Epsilon    = 9;   // Leerer Knoten (leere Alternative, leere Klammerung oder fehlerhafte Grammatik)

    /// <summary>
    /// Typ des Knotens
    /// </summary>
    internal readonly int type;

    /// <summary>
    /// Text des Knotens
    /// Dieses Feld speichert den direkt kopiert Text in den Knoten
    /// Action, Attribute und Resolver
    /// </summary>
    internal readonly string text;

    /// <summary>
    /// Token, das den Knoten reprsentiert.
    /// Production: ident
    /// Symbol: ident | string | char
    /// </summary>
    internal readonly Token sym;

    /// <summary>
    /// Separator-Zeichen fr eine Alternative '|'
    /// </summary>
    internal Token sep;

    /// <summary>
    /// ffnendes Token.
    /// Log: erstes Token (Prolog: null)
    /// Production: '='-Token
    /// Group: ffnende Klammer
    /// Action, Attribute, Resolver: erstes Token
    /// </summary>
    internal Token open;

    /// <summary>
    /// Schlieendes Token.
    /// Log: letztes Token
    /// Production: '.'-Token
    /// Sequence, Option, Iteration: Schlieende Klammer
    /// Action, Attribute, Resolver: letztes Token
    /// </summary>
    internal Token close;

    /// <summary>
    /// Zeiger auf ersten Knoten in Substruktur. 
    /// </summary>
    internal Node sub;

    /// <summary>
    /// Zeiger auf nchsten Knoten innerhalb der Struktur.
    /// </summary>
    internal Node next;
    
    /// <summary>
    /// Zeiger auf alternativen Knoten.
    /// </summary>
    internal Node alt;

    /// <summary>
    /// Grammar: Prolog Epilog
    /// Production: [Attribute] [Action]    
    /// Symbol: [Attribute]
    /// </summary>
    internal Node meta;

    /// <summary>
    /// Tiefe des Knotens im Syntaxgraphen.
    /// Wird nur gesetzt, falls nicht die gesamte ATG, sondern ein Knoten ausgegeben wird.
    /// </summary>
    internal int level;

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="type">Typ des Knotens (eine der Typkonstanten)</param>
    internal Node(int type)
    {
      this.type = type;
    }

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="type">Typ des Knotens (eine der Typkonstanten)</param>
    /// <param name="text"></param>
    internal Node(int type, string text) : this(type)
    {
      this.text = text;
    }

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="type">Typ des Knotens (eine der Typkonstanten)</param>
    /// <param name="text">Text des Knotens</param>
    /// <param name="open">ffnendes Token</param>
    /// <param name="close">Schlieendes Token</param>
    internal Node(int type, string text, Token open, Token close) : this(type,text)
    {
      this.open = open;
      this.close = close;
    }

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="type">Typ des Knotens (eine der Typkonstanten)</param>
    /// <param name="sym">Token</param>
    internal Node(int type, Token sym) : this(type,sym.val)
    {
      this.sym = sym;
    }

    /// <summary>
    /// Hngt this ans Ende der verketteten Liste.
    /// </summary>
    /// <param name="head">Kopf der Liste</param>
    /// <param name="curr">Letztes Listenelement</param>
    internal void Enqueue(Node head, ref Node curr)
    {
      curr = (curr == null) ? (head.next = this) : (curr.next = this);
    }

    /// <summary>
    /// Hngt this ans Ende der verketteten Liste von Alternativen.
    /// </summary>
    /// <param name="head">Kopf der Liste</param>
    /// <param name="curr">Letztes Listenelement</param>
    internal void EnqueueAlt(Node head, ref Node curr)
    {
      curr = (curr == null) ? (head.alt = this) : (curr.alt = this);
    }

    /// <summary>
    /// Hilfsmethode: Setzt den Level des Knoten.
    /// </summary>
    /// <param name="n">Knoten</param>
    /// <returns>n; falls n nicht null ist, wird dessen Level eingestellt.</returns>
    protected Node GetNode(Node n)
    {
      if (n != null)
      {
        n.level = level + 1;
      }
      return n;
    }

    /// <summary>
    /// Indexer. Gibt das Kind mit gegebenem Index zurck.
    /// </summary>
    public Node this[int symindex]
    {
      get
      {
        return this[0,symindex];
      }
    }

    /// <summary>
    /// Gibt den Knoten mit symindex in der Alternative altindex zurck.
    /// Indizes sind 0-basiert; falls keine Alternativen vorkommen, gilt altindex 0.
    /// </summary>
    public Node this[int altindex, int symindex]
    {
      get
      {
        Node n = sub;
        while ((altindex-- > 0) && (n != null))
        {
          n = n.alt;
        }
        while ((symindex-- > 0) && (n != null))
        {
          n = n.next;
        }
        return GetNode(n);
      }
    }
  }
}
