using System;
using System.IO;
using System.Text;

namespace Printy.Library
{
  /// <summary>
	/// Datenstruktur fr die zeilenweise Speicherung der formatierten Grammatik.
	/// Sie entspricht einer linearen Listen von Zeilen (Line-Objekten).
	/// Diese Klasse kmmert sich um Low-Level-Formatierungen wie Einzug,
	/// Spaltenbreite und Zeilenumbrche.
	/// </summary>
	public class TextBuilder
	{
    /// <summary>
    /// Schreibt count Zeichen in den StringBuilder.
    /// </summary>
    /// <param name="builder">Ziel</param>
    /// <param name="ch">Zeichen</param>
    /// <param name="count">Anzahl</param>
    static void AppendChars(StringBuilder builder, char ch, int count)
    {
      while (count-- > 0)
      {
        builder.Append(ch);
      }
    }

    /// <summary>
    /// Schreibt count Zeichen in den TextWriter.
    /// </summary>
    /// <param name="writer">Ziel</param>
    /// <param name="ch">Zeichen</param>
    /// <param name="count">Anzahl</param>
    static void WriteChars(TextWriter writer, char ch, int count)
    {
      while (count-- > 0)
      {
        writer.Write(ch);
      }
    }

    /// <summary>
    /// Grammar-Knoten
    /// </summary>
    protected internal readonly Grammar grammar;

    /// <summary>
    /// Konfiguration
    /// </summary>
    protected internal readonly Config config;

    /// <summary>
    /// Zeichenpuffer fr den linken Teil der aktuellen Zeile
    /// </summary>
    readonly StringBuilder builder;

    /// <summary>
    /// Dieser Stack wird fr die Verwaltung von Einzgen verwendet.
    /// Siehe Indent()
    /// </summary>
    readonly IntStack indentStack;

    /// <summary>
    /// Erste Zeile (Kopf der Liste)
    /// </summary>
    readonly Line head;

    /// <summary>
    /// Aktuelle Zeile (Schwanz der Liste)
    /// </summary>
    Line curr;

    /// <summary>
    /// Zuletzt geschriebenes Symbol.
    /// </summary>
    int sym;

    /// <summary>
    /// Tiefe von Klammerungen in der aktuellen Zeile. Dies wird fr den Blockstil bentigt.
    /// </summary>
    int par;

    /// <summary>
    /// Tatschliche Breite der linken Spalte (der lngsten Zeile der linken Spalte).
    /// </summary>
    int width;

    /// <summary>
    /// Anzahl semantischer Aktionen in aktueller Zeile (0 oder 1)
    /// </summary>
    int sem;

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="grammar">Grammar-Knoten</param>
    public TextBuilder(Grammar grammar) : this(grammar,new Config())
    {}

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="grammar">Grammar-Knoten</param>
    /// <param name="config">Config-Objekt</param>
    public TextBuilder(Grammar grammar, Config config)
		{
      this.grammar = grammar;
      this.config = config;
      builder = new StringBuilder();
      indentStack = new IntStack();
      head = curr = new Line();
		}

    /// <summary>
    /// Die tatschliche Breite der linken Spalte (der lngsten Zeile der linken Spalte).
    /// Diese kann die in der Konfiguration angegebene Spaltenbreite dann bersteigen,
    /// wann ein extra-langes Symbol vorkommt.
    /// Falls die berschneidung von Spalten erlaubt ist, gibt diese Eigenschaft die
    /// Breite der lngsten Zeile der linken Spalte mit semantischer Aktion an.
    /// </summary>
    public int Width
    {
      get
      {
        return width;
      }
    }

    /// <summary>
    /// Schliet die aktuelle Zeile. Diese Methode muss aufgerufen werden, wenn eine Zeile
    /// fertig beschrieben worden ist.
    /// </summary>
    public void CloseLine()
    {
      if (builder.Length > 0)
      {
        curr.left = builder.ToString();
        builder.Remove(0,builder.Length);
      }
      if ((curr.Width > width) && ((sem > 0) || !config.overlapSet[sym])) // Spaltenbreite erneuern
      {
        width = curr.Width;
      }
    }

    /// <summary>
    /// Fgt eine neue Zeile ein.
    /// </summary>
    public void NewLine()
    {
      curr = curr.next = new Line();
      curr.indent = indentStack.Top();
      par = sym = sem = 0;
    }

    /// <summary>
    /// Setzt den Text fr die semantische Aktion der aktuellen Zeile.
    /// </summary>
    /// <param name="action">Text der semantischen Aktion</param>
    public void PutAction(string action)
    {
      sem++;
      curr.right = Preprocess(action,Config.Action);
    }

    /// <summary>
    /// Prft, ob der angegebene Text in die aktuelle Zeile geschrieben werden kann.
    /// Implementierungsanmerkung zu den Zeilen "return config.openSet[sym];"
    ///   Dies hat den Effekt, dass bei berschreiten der Einzugs- oder Spaltenbreite
    ///   nur dann ein Zeilenwechsel erfolgt, falls nicht direkt davor ein ffnendes Symbol
    ///   steht. Wrde dort ein Zeilenwechsel erfolgen, wre das unschne Ergebnis, dass
    ///   die Spaltenbreite trotzdem berschritten wird; also wird der Zeilenwechsel a priori
    ///   unterdrckt.
    /// </summary>
    /// <param name="text">Der Text</param>
    /// <param name="type">Typ des Textes</param>
    /// <param name="level">Ebene</param>
    /// <returns>
    /// true in folgenden Fllen:
    /// - die aktuelle Zeile ist leer (dann knnen auch extra-lange Symbolen aufgenommen werden)
    /// - die Konfiguration verbietet einen Zeilenwechsel (durch Hard Breaks)
    /// - der Text hat tatschlich Platz in der aktuellen Zeile (Standardfall)
    /// </returns>
    public bool CanAppendText(string text, int type, int level)
    {
      if (sym == 0)
      {
        return true; // leere Zeile kann immer befllt werden
      }
      int breakLevel = config.GetBreak(sym,type);
      /* --- Hard Breaks --- */
      if (breakLevel == Config.BREAK_NEVER)
      {
        return true; // Zeilenwechsel wird unterdrckt, Anhngen von Text immer mglich
      }
      if (Config.TestLevel(level,breakLevel))
      {
        return false; // Zeilenwechsel wird erzwungen, Anhngen von Text nicht mglich
      }
      /* --- Soft Breaks --- */
      if ((par == 0) && config.autoBlock && config.closeSet[type])
      {
        return false; // Auto-Blockstil => schlieendes Symbol in nchste Zeile
      }
      int space = config.GetSpace(sym,type);
      if (config.openSet[type] && (curr.Width + space > config.indentWidth))
      {
        return config.openSet[sym]; // Einzugsbreite berschritten => ffnendes Symbol in nchste Zeile
      }
      if (curr.Width + space + text.Length > config.colWidth)
      {
        return config.openSet[sym]; // Spaltenbreite berschritten => Symbol in nchste Zeile
      }
      /* Text hat in aktueller Zeile Platz */
      return true;
    }

    /// <summary>
    /// Diese Hilfsmethode verwaltet die Einrckungen.
    /// Jedes Mal, wenn ein ffnendes Symbol gedruckt wird, wird die aktuelle
    /// Spalte auf den Stapel gelegt.
    /// Bei einem schlieenden Symbol wird ein Element heruntergenommen.
    /// </summary>
    /// <param name="type">Typ des aktuellen Knoten</param>
    void IndentBlock(int type)
    {
      if (config.openSet[type])
      {
        indentStack.Push(curr.Width);
        par++;
      }
      else if (config.closeSet[type])
      {
        indentStack.Pop();
        if (par > 0)
        {
          par--;
        }
      }
    }

    /// <summary>
    /// Prprozessor-Methode, die fr jeden einzelnen Text augerufen wird.
    /// </summary>
    /// <param name="text">Text</param>
    /// <param name="type">Typ des Textes</param>
    /// <returns>Den Text nach dem Prprozessor, standardmig denselben Text</returns>
    protected virtual string Preprocess(string text, int type)
    {
      return text;
    }

    /// <summary>
    /// Fgt den Text in die aktuelle Zeile ein.
    /// </summary>
    /// <param name="text">Der Text</param>
    /// <param name="type">Typ des Textes</param>
    public void AppendText(string text, int type)
    {
      int space = config.GetSpace(sym,type);
      AppendChars(builder,config.spaceChar,space);
      curr.length += space;
      IndentBlock(type);
      builder.Append(curr.left = Preprocess(text,type));
      curr.length += text.Length;
      sym = type;
    }

    /// <summary>
    /// Gibt den Textpuffer in den TextWriter aus.
    /// </summary>
    /// <param name="writer">Ein TextWriter Objekt, das nicht null sein darf</param>
    public virtual void WriteContentTo(TextWriter writer)
    {
      WriteBufferTo(writer);
    }

    /// <summary>
    /// Gibt den Textpuffer in den TextWriter aus.
    /// </summary>
    /// <param name="writer">Ein TextWriter Objekt, das nicht null sein darf</param>
    public void WriteBufferTo(TextWriter writer)
    {
      int lineNo = 0;
      int initialSpace = config.colSpace + (config.autoFit ? width : Math.Max(width,config.colWidth));
      char spaceChar = config.spaceChar;
      string lineNum = config.lineNum;

      for (Line line = head; line != null; line = line.next)
      {
        int space = initialSpace;
        if (lineNum != null)
        {
          writer.Write(Preprocess((++lineNo).ToString(lineNum),Config.LineNumber));
        }
        if (line.left != null)
        {
          WriteChars(writer,spaceChar,line.indent);
          writer.Write(line.left);
          space -= line.Width;
        }
        if (line.right != null)
        {
          WriteChars(writer,spaceChar,space);
          writer.Write(line.right);
        }
        writer.WriteLine();
      }
    }
  }

  /// <summary>
  /// Ausgabezeile mit Grammatik links und Semantik rechts.
  /// </summary>
  class Line
  {
    /// <summary>
    /// Einzug
    /// </summary>
    internal int indent;

    /// <summary>
    /// Lnge des linken Textes (ohne Einzug)
    /// Der Wert muss nicht zwingend gleich left.Length sein, da z.B. HTML-Escapes in length
    /// nicht mitgezhlt werden (siehe Preprocess)
    /// </summary>
    internal int length;

    /// <summary>
    /// Linker Text
    /// </summary>
    internal string left;

    /// <summary>
    /// Rechter Text
    /// </summary>
    internal string right;

    /// <summary>
    /// Nchste Zeile
    /// </summary>
    internal Line next;

    /// <summary>
    /// Breite der linken Spalte.
    /// </summary>
    internal int Width
    {
      get
      {
        return indent + length;
      }
    }
  }

  /// <summary>
  /// Ein reduzierter Stack fr den Datentyp int.
  /// Die Methoden werfen keine Exceptions sondern geben 0 zurck, falls der Stack leer ist.
  /// </summary>
  class IntStack
  {
    /// <summary>
    /// Daten
    /// </summary>
    int[] data = new int[8]; 

    /// <summary>
    /// Gre
    /// </summary>
    int count;

    /// <summary>
    /// Legt das Element auf den Stack
    /// </summary>
    /// <param name="val">Wert des Elements</param>
    internal void Push(int val)
    {
      if (count == data.Length) // Array vergrern
      {
        int[] newData = new int[data.Length << 1];
        Array.Copy(data,newData,count);
        data = newData;
      }
      data[count++] = val;
    }

    /// <summary>
    /// Entfernt des oberste Element vom Stapel.
    /// </summary>
    internal void Pop()
    {
      if (count > 0)
      {
        count--;
      }
    }

    /// <summary>
    /// Gibt das oberste Element zurck, oder 0.
    /// </summary>
    /// <returns>Oberstes Element des Stapels oder 0, falls er leer ist.</returns>
    internal int Top()
    {
      return (count == 0) ? 0 : data[count - 1];
    }
  }
}
