using System;
using System.Collections;
using System.IO;
using System.Reflection;
using Printy.Library;

using System.Text;

namespace Printy.Tool
{
	/// <summary>
	/// Konsolenschnittstelle
	/// </summary>
	public class Tool
	{
    /// <summary>
    /// Der Haupteinstiegspunkt fr die Anwendung.
    /// </summary>
    /// <param name="argv">Parameter-Array</param>
    [STAThread]
    public static void Main(params string[] argv)
    {
      try
      {
        Execute(argv);
      }
      catch (IOException ex) // Datei konnte nicht geffnet werden...
      {
        Error(ex.Message);
      }
    }

    /// <summary>
    /// Fhrt den Printy aus.
    /// Diese Methode wirft im Fehler Fall eine IOException.
    /// </summary>
    /// <param name="argv">Parameter-Array</param>
    public static void Execute(params string[] argv)
    {
      if (argv.Length < 2) // Hilfedatei (Synopsis.txt) aus Assembly (Printy.exe) laden
      {
        Assembly assembly = Assembly.GetExecutingAssembly();
        Stream stream = assembly.GetManifestResourceStream("Synopsis.txt");
        TextReader reader = new StreamReader(stream);
        Console.Write(reader.ReadToEnd());
        return;
      }
      Printy.Library.Parser parser = new Printy.Library.Parser(new Printy.Library.Scanner(argv[0]));
      parser.Parse();
      if (parser.errors.count > 0) 
      {
        Error("{0} errors detected.",parser.errors.count);
      }

      StreamWriter writer;
      TextBuilder builder;
      Grammar grammar = parser.grammar;
      Node node = grammar;
      string content = argv[1];
      string ext = GetExtension(content);
      Settings s = new Settings();
      s.config = new Config();
      s.html = ext == "html";
      s.ebnf = ext == "ebnf";
      
      if (argv.Length > 2) // Parameter parsen
      {
        ParamSet pset = new ParamSet();
        Param param = null;

        for (int i = 2; i < argv.Length; i++)
        {
          String arg = argv[i];
          if (arg.StartsWith("-") || arg.StartsWith("/"))
          {
            Param p = (Param)Param.table[arg.Substring(1)];
            if (p != null)
            {
              CheckArg(param);
              param = p;
              p.pset = true;
              continue;
            }
          }
          if (param != null)
          {
            param.arg = arg;
          }
        }
        CheckArg(param);
        pset.Apply(s);
      }

      if (s.html) // HyperConfig anpassen und HyperTextBuilder verwenden
      {
        s.htmlConfig = new HyperConfig();
        HyperTextBuilder htmlBuilder = new HyperTextBuilder(grammar,s.config,s.htmlConfig);

        if (s.name != null)
        {
          s.htmlConfig.createLinks = false;
        }
        if (s.style != null)
        {
          s.htmlConfig.styleSheet = s.style;
          s.htmlConfig.styleSet |= true;
        }
        if (s.overview != null)
        {
          if (s.index != null)
          {
            writer = new StreamWriter(s.index);
            htmlBuilder.WriteIndexTo(writer,s.overview,content,true);
            writer.Close();
          }
          writer = new StreamWriter(s.overview);
          htmlBuilder.WriteOverviewTo(writer,content,true);
          writer.Close();
        }
        builder = htmlBuilder;
      }
      else // TextBuilder verwenden
      {
        builder = new TextBuilder(grammar,s.config);
      }
      if (s.ebnf)
      {
        s.config.hideSet = 
          new BitSet(Config.Attribute,Config.Resolver,Config.Action,Config.Prolog,Config.Epilog);
      }
      if (s.name != null)
      {
        node = grammar[s.name];
        if (node == null)
        {
          Error("Production '{0}' not found.",s.name);
        }
      }
      if (s.configFile != null)
      {
        Printy.Tool.Parser p = new Printy.Tool.Parser(new Printy.Tool.Scanner(s.configFile));
        p.config = s.config;
        p.htmlConfig = s.htmlConfig;
        p.Parse();
        if (p.errors.count > 0)
        {
          Error("{0} config errors detected.",p.errors.count);
        }
      }

      new Printer().Print(builder,node);
      writer = new StreamWriter(content);
      builder.WriteContentTo(writer);
      writer.Close();
    }

    /// <summary>
    /// Gibt die Dateinamenserweiterung zurck.
    /// </summary>
    /// <param name="filename">Dateiname</param>
    /// <returns>Dateiendung (Text nach letztem '.', eklusive, oder null,
    /// falls der Dateiname keine spezifische Endung hat.
    /// </returns>
    static string GetExtension(string filename)
    {
      int i = filename.LastIndexOf('.');
      return (i >= 0) ? filename.Substring(i + 1).ToLower() : null;
    }

    /// <summary>
    /// berprft, ob der Parameter gltige Argumente hat.
    /// </summary>
    /// <param name="param">Param-Objekt (darf null sein)</param>
    static void CheckArg(Param param)
    {
      if ((param != null) && param.pset && param.parg && (param.arg == null))
      {
        Error("Parameter '{0}' requires argument.",param.name);
      }
    }

    /// <summary>
    /// Gibt die Fehlermeldung aus und beendet das Programm.
    /// </summary>
    /// <param name="format">Formatstring</param>
    /// <param name="argv">Parameter-Array</param>
    static void Error(string format, params object[] argv)
    {
      Console.WriteLine(format,argv);
      Environment.Exit(1);
    }
	}

  /// <summary>
  /// Einstellungen
  /// </summary>
  class Settings
  {
    /// <summary>
    /// Konfiguration
    /// </summary>
    internal Config config;

    /// <summary>
    /// Html-Konfiguration
    /// </summary>
    internal HyperConfig htmlConfig;

    /// <summary>
    /// config.txt
    /// </summary>
    internal string configFile;

    /// <summary>
    /// index.html
    /// </summary>
    internal string index;

    /// <summary>
    /// overview.html
    /// </summary>
    internal string overview;

    /// <summary>
    /// style.css
    /// </summary>
    internal string style;

    /// <summary>
    /// Name der auszugebenden Produktion
    /// </summary>
    internal string name;

    /// <summary>
    /// HTML-Ausgabe?
    /// </summary>
    internal bool html;

    /// <summary>
    /// EBNF-Ausgabe?
    /// </summary>
    internal bool ebnf;
  }

  /// <summary>
  /// Klasse fr die Gesamtheit aller Parameter.
  /// </summary>
  class ParamSet
  {
    /// <summary>
    /// Parameter-Objekte
    /// </summary>
    internal readonly Param
      config       = new Param("c","config",false),
      index        = new Param("i","index",false),
      overview     = new Param("o","overview",false),
      style        = new Param("s","style",false),
      name         = new Param("n","name",true),
      html         = new Param("h","html",false),
      ebnf         = new Param("e","ebnf",false),
      altPar       = new Param("ap","altPar",false),
      indentWidth  = new Param("iw","indentWidth",true),
      colWidth     = new Param("cw","colWidth",true),
      colSpace     = new Param("cs","colSpace",true),
      autoBlock    = new Param("ab","autoBlock",false),
      autoFit      = new Param("af","autoFit",false),
      noBreaks     = new Param("nb","noBreaks",false),
      spaceChar    = new Param("sc","spaceChar",true),
      lineNum      = new Param("ln","lineNum",true),
      altParOpen   = new Param("apo","altParOpen",true),
      altParClose  = new Param("apc","altParClose",true),
      colOverlap   = new Param("co","colOverlap",false),
      brProduction = new Param("bp","brProduction",false),
      brAssign     = new Param("ba","brAssign",false),
      brDelimiter  = new Param("bd","brDelimiter",false),
      brSeparator  = new Param("bs","brSeparator",false),
      brOpen       = new Param("bo","brOpen",false),
      brClose      = new Param("bc","brClose",false);
  
    /// <summary>
    /// Setzt die Parameter in den Einstellungen um.
    /// </summary>
    /// <param name="s">Settings-Objekt</param>
    internal void Apply(Settings s)
    {
      if (config.pset)
      {
        s.configFile = config.StringValue("config.txt");
      }
      if (index.pset)
      {
        s.index = index.StringValue("index.html");
      }
      if (overview.pset)
      {
        s.overview = overview.StringValue("overview.html");
      }
      if (style.pset)
      {
        s.style = style.StringValue("style.css");
      }
      if (name.pset)
      {
        s.name = name.StringValue(null);
      }
      if (html.pset)
      {
        s.html = html.BoolValue(true);
      }
      if (ebnf.pset)
      {
        s.ebnf = ebnf.BoolValue(true);
      }
      if (altPar.pset)
      {
        s.config.altPar = altPar.IntValue(Config.ALT_PAR_ALWAYS);
      }
      if (indentWidth.pset)
      {
        s.config.indentWidth = indentWidth.IntValue(0);
      }
      if (colWidth.pset)      
      {
        s.config.colWidth = colWidth.IntValue(0);
      }
      if (colSpace.pset)      
      {
        s.config.colSpace = colSpace.IntValue(0);
      }
      if (autoBlock.pset)    
      {
        s.config.autoBlock = autoBlock.BoolValue(true);
      }
      if (autoFit.pset)        
      {
        s.config.autoFit = autoFit.BoolValue(true);
      }
      if (noBreaks.pset)
      {
        s.config.noBreaks = noBreaks.BoolValue(true);
      }
      if (spaceChar.pset)
      {
        s.config.spaceChar = spaceChar.CharValue(' ');
      }
      if (lineNum.pset)
      {
        s.config.lineNum = lineNum.StringValue(null);
      }
      if (altParOpen.pset)
      {
        s.config.altParOpen = altParOpen.StringValue(null);
      }
      if (altParClose.pset)
      {
        s.config.altParClose = altParClose.StringValue(null);
      }
      if (colOverlap.pset) 
      {
        s.config.overlapSet |= colOverlap.BoolValue(true);
      }
      if (brProduction.pset)
      {
        int breakLevel = brProduction.BoolValue(true) ? Config.BREAK_ALWAYS : Config.BREAK_DEFAULT;
        s.config.PutBreak(Config.ALL,Config.Production,breakLevel);
      }
      if (brAssign.pset)
      {
        int breakLevel = brAssign.BoolValue(true) ? Config.BREAK_ALWAYS : Config.BREAK_DEFAULT;
        s.config.PutBreak(Config.ALL,Config.Assign,breakLevel);
      }
      if (brDelimiter.pset)
      {
        int breakLevel = brDelimiter.BoolValue(true) ? Config.BREAK_ALWAYS : Config.BREAK_DEFAULT;
        s.config.PutBreak(Config.ALL,Config.Delimiter,breakLevel);
      }
      if (brSeparator.pset)
      {
        int breakLevel = brSeparator.IntValue(Config.BREAK_ALWAYS);
        s.config.PutBreak(Config.ALL,Config.Separator,breakLevel);
      }
      if (brOpen.pset)
      {
        int breakLevel = brOpen.IntValue(Config.BREAK_ALWAYS);
        s.config.PutBreak(Config.ALL,Config.Open,breakLevel);
      }
      if (brClose.pset)
      {
        int breakLevel = brClose.IntValue(Config.BREAK_ALWAYS);
        s.config.PutBreak(Config.ALL,Config.Close,breakLevel);
      }
    }
  }

  /// <summary>
  /// Klasse fr Parameter nach dem typesafe enum pattern
  /// </summary>
  class Param
  {
    /// <summary>
    /// Abbildung string -> Param
    /// </summary>
    internal static readonly Hashtable table = new Hashtable();

    /// <summary>
    /// Konstruktor
    /// </summary>
    /// <param name="code">Krzel fr den Parameter</param>
    /// <param name="name">Name des Parameters</param>
    /// <param name="parg">Argument erforderlich?</param>
    internal Param(string code, string name, bool parg)
    {
      table[code] = this;
      table[name] = this;
      this.name = name;
      this.parg = parg;
    }

    /// <summary>
    /// Name des Parameters
    /// </summary>
    internal readonly string name;

    /// <summary>
    /// Argument notwendig?
    /// </summary>
    internal readonly bool parg;

    /// <summary>
    /// true, falls der Parameter gesetzt ist
    /// </summary>
    internal bool pset;

    /// <summary>
    /// Wert des Arguments
    /// </summary>
    internal string arg;

    /// <summary>
    /// Interpretiert das Zeichenkette als Zahl und gibt dessen Wert zurck.
    /// </summary>
    /// <returns></returns>
    internal string StringValue(string defaultValue)
    {
      return (arg == null) ? defaultValue : arg;
    }

    /// <summary>
    /// Interpretiert das Argument als Zeichen als Zahl und gibt dessen Wert zurck.
    /// </summary>
    /// <param name="defaultValue">Standardwert</param>
    /// <returns>Wert des Arguments</returns>
    internal char CharValue(char defaultValue)
    {
      return ((arg == null) || (arg.Length != 1)) ? defaultValue : arg[0];
    }

    /// <summary>
    /// Interpretiert das Argument als Wahrheitswert als Zahl und gibt dessen Wert zurck.
    /// </summary>
    /// <param name="defaultValue">Standardwert</param>
    /// <returns></returns>
    internal bool BoolValue(bool defaultValue)
    {
      return (arg == null) ? defaultValue : (arg.Equals("+") || arg.ToLower().Equals("true"));
    }

    /// <summary>
    /// Interpretiert das Argument als ganze Zahl und gibt dessen Wert zurck.
    /// </summary>
    /// <param name="defaultValue">Standardwert</param>
    /// <returns>Wert des Arguments</returns>
    internal int IntValue(int defaultValue)
    {
      return (arg == null) ? defaultValue : ScanIntValue();
    }

    /// <summary>
    /// Hilfsmethode fr IntValue: scanned einen int aus dem Argument-String.
    /// </summary>
    /// <returns>Text als Zahl interpretiert</returns>
    int ScanIntValue()
    {
      string text = arg;
      int result = 0;
      bool neg = false;

      for (int i = 0; i < text.Length; i++)
      {
        int ch = text[i];
        if (ch == '-')
        {
          neg = !neg;
        }
        else if (('0' <= ch) && (ch <= '9'))
        {
          result = result * 10 + ch - '0';
        }
        else
        {
          break;
        }
      }

      return neg ? -result : result;
    }
  }
}
